////////////////////////////////////////////////////////////////////////////////
//Descripion:   4 bit parallel LCD driver                                      /
//              Derived from lcd4bit_old.c, which required the data bits to be /
//              the bottom of a port and fast I/O to be set up so that other   /
//              bits in the port were not affected.  This version uses only bit/
//              wise port accesses via the actual memory addresses of the bits /
//              (no input/output calls.   Not only does this make the code more/
//              transportable to other processors but now allows all 7 bits of /
//              the LCD interface to be randomly placed on available pins.     /
////////////////////////////////////////////////////////////////////////////////
//Taylor Electronics Services (TES)                                            /
//                                                                             /
//  This source code is public domain.                                         /
////////////////////////////////////////////////////////////////////////////////
//Revision history                                                             /
//No.   Date        Description                                                /
//-----------------------------------------------------------------------------/
//00    2005/10/23  Initial release                                            /
//01    2009/09/14  Added stand alone busy test for buffering routine calls    /
////////////////////////////////////////////////////////////////////////////////
//printf(lcd_char,"formatting stuff",variables);                               /
////////////////////////////////////////////////////////////////////////////////
//lcd_init()            initialize display                                     /
//lcd_busy()            Return 1 if busy, 0 if not busy                        /    
//lcd_char(data)        xxxx xxxx   Ascii data to write to current pointer     /
//                                  Pointer:0b01000000-0b01111111 = cgen ram   /
//                                          0b10000000-0b11111111 = disp ram   /
//                                                                             /
//lcd_cmmd(data)        xxxx xxxx   instruction data:                          /
//                      0000 0001   LCD clear display                          /
//                      0000 0010   Set cursor to home                         /
//                      0000 01ds   d0:  decrement position with each character/
//                                  d1:  increment position                    /
//                                  s0:  no display shift with each character  /
//                                  s1:  display shifts                        /
//                      0000 1dcb   d0:  display off                           /
//                                  d1:  display on                            /
//                                  c0:  cursor off                            /
//                                  c1:  cursor on                             /
//                                  b0:  cursor blink off                      /
//                                  b1:  cursor blink on                       /
//                      0001 sr00   s0:  cursor moves with character writes    /
//                                  s1:  display shifts with character writes  /
//                                  r0:  shift left                            /
//                                  r1:  shift right                           /
//                      001d nf00   d0:  data width is 4 bits                  /
//                                  d1:  data width is 8 bits >>DON't USE!<<   /
//                                  n0:  1 line               >>DON't USE!<<   /
//                                  n1:  2 lines                               /
//                                  f0:  font is 5x7                           /
//                                  f1:  xxx                  >>DON't USE!<<   /
//                      01aa aaaa   a.a  character generator ram address set   /
//                                       character data sent after this and    /
//                                       until a display address is set will be/
//                                       written to the character generator ram/
//                      1aaa aaaa   a.a  display address ram address set       /
//                                       aka 'Cursor' set                      /
//                                       0x80:  line 1 position 0              /
//                                       0xC0:  line 2 position 0              /
//                                       character data sent after this and    /
//                                       until a character generator address is/
//                                       set will be written to the display ram/
////////////////////////////////////////////////////////////////////////////////
//include in main c file
//for paged ram devices, use full address ie [0x05.p for port]/[0x85.p for tris]
//for devices with both a latch and a pin register, use the latch registers
//for RS,RW and EN, but use the pin registers for D0-D3
//
//example code below is for an 18F252 type (has latch registers)
/*
#include    "pinmap.h"                  //name conversions
//                                      //(1)  Gnd
//                                      //(2)  Vdd
//                                      //(3)  Contrast
#define     lcd_rs          a3lat       //(4)  Control registers
#define     lcd_rw          a2lat       //(5)
#define     lcd_en          a1lat       //(6)
//                                      //(7)  Gnd
//                                      //(8)  Gnd
//                                      //(9)  Gnd
//                                      //(10) Gnd
#define     lcd_d0          b2pin       //(11) Data registers
#define     lcd_d1          b3pin       //(12)
#define     lcd_d2          b4pin       //(13)
#define     lcd_d3          a0pin       //(14)
#define     lcd_rs_dir      a3dir       //control direction registers
#define     lcd_rw_dir      a2dir       
#define     lcd_en_dir      a1dir
#define     lcd_d0_dir      b2dir       //data direction registers
#define     lcd_d1_dir      b3dir
#define     lcd_d2_dir      b4dir
#define     lcd_d3_dir      a0dir
#include    "lcd4bit.c"
*/
////////////////////////////////////////////////////////////////////////////////
#define lcd_line_1  0x80                //to set the cursor to line1, char 18:
#define lcd_line1   0x80                //to set the cursor to line1, char 18:
#define lcd_line_2  0xc0                //    lcd_cmmd(lcd_line_1 | 18);
#define lcd_line2   0xc0                //    lcd_cmmd(lcd_line_1 | 18);
#define lcd_cls     0x01                //clear display
#define lcd_home    0x02                //set cursor to home (cursor 0x00)
////////////////////////////////////////////////////////////////////////////////
//using this global saves 18 bytes of rom (9 instructions)
int8 lcd_writemode_speed;               //init vs fast delay message global
////////////////////////////////////////////////////////////////////////////////
int8 lcd_clock(void){
int8 temp=0;
lcd_en=1;                               //enable high
delay_cycles(3);                        //min pw-hi=230ns
    if(lcd_d0)                          //bitwise port read
        temp=temp|1;
    if(lcd_d1)
        temp=temp|2;
    if(lcd_d2)
        temp=temp|4;
    if(lcd_d3)
        temp=temp|8;
lcd_en=0;                               //enable low
delay_cycles(3);                        //min pw-low=270ns
return temp;
}
////////////////////////////////////////////////////////////////////////////////
//write data to port, bitwise
//mode=0: fast, mode=1: slow
void lcd_data_wr(int8 dataout){
lcd_rw=0;                               //set for write
lcd_d0_dir=0;                           //set data to output
lcd_d1_dir=0;
lcd_d2_dir=0;
lcd_d3_dir=0;
//..............................................................................
lcd_d0=bit_test(dataout,4);             //write msb's
lcd_d1=bit_test(dataout,5);
lcd_d2=bit_test(dataout,6);
lcd_d3=bit_test(dataout,7);
//..............................................................................
lcd_clock();                            //clock out msb's
if (lcd_writemode_speed==1)             //long delay for init
    delay_ms(10);
//..............................................................................
lcd_d0=bit_test(dataout,0);             //write lsb's
lcd_d1=bit_test(dataout,1);
lcd_d2=bit_test(dataout,2);
lcd_d3=bit_test(dataout,3);
//..............................................................................
lcd_clock();                            //clock out lsb's
if (lcd_writemode_speed==1)             //long delay for init
    delay_ms(10);
return;
}
////////////////////////////////////////////////////////////////////////////////
//initialize display                                                           /
////////////////////////////////////////////////////////////////////////////////
void lcd_init(){
lcd_writemode_speed=1;                  //slow write
lcd_rw_dir=0;                           //set control pins as output
lcd_rs_dir=0;
lcd_en_dir=0;
//..............................................................................
lcd_en=0;                               //set control state for initialization
delay_cycles(1);
lcd_rs=0;
delay_cycles(1);
lcd_rw=0;
delay_ms(50);
//..............................................................................
lcd_data_wr(0x33);                      //this saves 4 bytes (2 instr) over
lcd_data_wr(0x32);                      //  a loop accessing an array
lcd_data_wr(0x28);
lcd_data_wr(0x06);
lcd_data_wr(0x0c);
lcd_data_wr(0x01);
//..............................................................................
return;
}
////////////////////////////////////////////////////////////////////////////////
int8 lcd_data_rd(){
int8 temp;
lcd_rw=1;                               //set for read
lcd_d0_dir=1;                           //set data to output
lcd_d1_dir=1;
lcd_d2_dir=1;
lcd_d3_dir=1;
temp=lcd_clock()<<4;                    //read msb's
temp=temp | lcd_clock();                //read lsb's
return temp;
}
////////////////////////////////////////////////////////////////////////////////
//read busy flag (0b1xxxxxxxx > 0b0xxxxxxx), return status
unsigned int1 lcd_busy(void){
lcd_rs=0;                               //0=command, in this case busy flag
return(bit_test(lcd_data_rd(),7));
}
////////////////////////////////////////////////////////////////////////////////
//read busy flag (0b1xxxxxxxx > 0b0xxxxxxx), return when not busy
void lcd_busyx(void){
lcd_rs=0;                               //0=command, in this case busy flag
while(bit_test(lcd_data_rd(),7));
return;
}
////////////////////////////////////////////////////////////////////////////////
//0b01xxxxxx    character generator address set
//
//0b1xxxxxxx    display matrix address set
//0b10xxxxxx    line 0 (64 characters)
//0b11xxxxxx    line 2 (64 characters)
void lcd_cmmd(dataout){
lcd_writemode_speed=0;                  //global-0 = fast write
lcd_busyx();                            //wait till not busy
lcd_rs=0;                               //1=command
lcd_data_wr(dataout);
return;
}
////////////////////////////////////////////////////////////////////////////////
void lcd_char(dataout){
lcd_writemode_speed=0;                  //global-0 = fast write
lcd_busyx();                            //wait till not busy
lcd_rs=1;                               //1=data
lcd_data_wr(dataout);
return;
}
////////////////////////////////////////////////////////////////////////////////


